import { FlacStreamTagger } from "flac-stream-tagger";
import Ffmpeg from "fluent-ffmpeg";
import ffmpegPath from "ffmpeg-static";
import { PassThrough } from "stream";
import { DashMPD } from '@liveinstantly/dash-mpd-parser';


Ffmpeg.setFfmpegPath(ffmpegPath as string);

export default class {
    authHeaders: { [key: string]: string }

    constructor(authHeaders: { [key: string]: string }) {
        this.authHeaders = authHeaders
    }

    async convertAacToFlac(buffer: Buffer<ArrayBuffer> | ArrayBuffer) {
        const inputStream = new PassThrough();
        inputStream.end(Buffer.isBuffer(buffer) ? buffer : Buffer.from(buffer));

        const outputChunks: Buffer[] = [];

        await new Promise((resolve, reject) => {
            Ffmpeg(inputStream)
                .format("flac")
                .on("error", reject)
                .on("end", resolve)
                .pipe()
                .on("data", chunk => outputChunks.push(chunk));
        });

        return Buffer.concat(outputChunks);;
    }

    async convertMp4ToFlac(buffer: Buffer) {
        const inputStream = new PassThrough();
        inputStream.end(buffer);

        const outputChunks: Buffer[] = [];

        await new Promise((resolve, reject) => {
            Ffmpeg(inputStream)
                .audioCodec("flac")
                .format("flac")
                .on("error", reject)
                .on("end", () => resolve(Buffer.concat(outputChunks)))
                .pipe()
                .on("data", chunk => outputChunks.push(chunk));
        });

        return Buffer.concat(outputChunks);
    }

    async fetchTrack(id: number) {
        const audio = await fetch(`https://desktop.tidal.com/v1/tracks/${id}/playbackinfo?audioquality=HI_RES_LOSSLESS&playbackmode=STREAM&assetpresentation=FULL`, {
            headers: this.authHeaders
        })

        console.log(await audio.text())

        return await audio.json() as {
            trackId: number,
            audioPresentation: string,
            audioMode: string,
            audioQuality: string,
            manifestMimeType: string,
            manifestHash: string,
            manifest: string,
            albumReplayGain: number,
            albumPeakAmplitude: number,
            trackReplayGain: number,
            trackPeakAmplitude: number,
            bitDepth: number,
            sampleRate: number,
            status?: number,
        }
    }

    async fetchTracks() {
        const tracks = await fetch("https://listen.tidal.com/v1/users/199235629/favorites/tracks?offset=0&limit=10000&order=DATE&orderDirection=DESC&countryCode=US", {
            headers: this.authHeaders
        })

        return await tracks.json();
    }

    async downloadFlac(manifestMimeType: string, manifest: string) {
        if (manifestMimeType === "application/dash+xml") {
            const mpd = new DashMPD();
            mpd.parse(Buffer.from(manifest, "base64").toString("utf-8"))

            const Representation = mpd.getJSON().MPD.Period[0].AdaptationSet[0].Representation[0];
            const segmentTemplate = Representation.SegmentTemplate;
            const segmentCount = segmentTemplate.SegmentTimeline.S[0]["@r"];
            const segmentBuffers = [];

            const initialization = await fetch(segmentTemplate["@initialization"])
            const initializationBuffer = await initialization.arrayBuffer()
            segmentBuffers.push(Buffer.from(initializationBuffer));

            const segmentPromises = [];
            for (let i = 0; i < 1 + segmentCount; i++) {
                const segmentUrl = segmentTemplate["@media"].replace("$Number$", segmentTemplate["@startNumber"] + i);
                segmentPromises.push(fetch(segmentUrl).then(segment => segment.arrayBuffer()).then(segmentBuffer => Buffer.from(segmentBuffer)));
            }

            const segmentBuffersArray = await Promise.all(segmentPromises);
            segmentBuffers.push(...segmentBuffersArray);

            const flac = await this.convertMp4ToFlac(Buffer.concat(segmentBuffers))

            return { buffer: flac, mimeType: "audio/flac" }
        } else if (manifestMimeType === "application/vnd.tidal.bts") {
            const data = JSON.parse(Buffer.from(manifest, "base64").toString("utf-8")) as {
                "mimeType": string,
                "codecs": string,
                "encryptionType": string,
                "urls": string[]
            }

            const flac = await fetch(data.urls[0])

            return { buffer: await flac.arrayBuffer(), mimeType: data.mimeType }
        }

        return { buffer: new ArrayBuffer(0), mimeType: "" } // TODO: Handle other mime types
    }

    async tagFlac(id: number, audioBuffer: { buffer: Buffer<ArrayBuffer> | ArrayBuffer, mimeType: string }) {
        const fileId = Bun.nanoseconds().toString(36);

        if (audioBuffer.mimeType === "audio/mp4") {
            audioBuffer.buffer = await this.convertAacToFlac(audioBuffer.buffer);
        }

        const trackReq = await fetch(`https://api.tidal.com/v1/tracks/${id}/?countryCode=US`, {
            headers: this.authHeaders
        })

        const track = await trackReq.json() as {
            title: string,
            replayGain: number,
            peak: number,
            streamStartDate: string,
            trackNumber: number,
            volumeNumber: number, // discNumber
            copyright: string,
            bpm: number,
            url: string, // comment
            isrc: string,
            artists: [
                {
                    name: string,
                }
            ],
            album: {
                id: number,
                title: string,
                cover: string,
            },
            status?: string,
        };

        const alubmReq = await fetch(`https://api.tidal.com/v1/albums/${track.album.id}/?countryCode=US`, {
            headers: this.authHeaders
        })

        const album = await alubmReq.json() as {
            title: string,
            numberOfTracks: number,
            releaseDate: string,
            cover: string,
            upc: string,
            artists: [
                {
                    name: string,
                }
            ],
        }

        return FlacStreamTagger.fromBuffer(Buffer.from(Buffer.isBuffer(audioBuffer.buffer) ? audioBuffer.buffer : Buffer.from(audioBuffer.buffer)), {
            tagMap: {
                title: track.title,
                trackNumber: track.trackNumber.toString(),
                discNumber: track.volumeNumber.toString(),
                bpm: (track.bpm || 0).toString(),
                date: track.streamStartDate,
                copyright: track.copyright,
                REPLAYGAIN_TRACK_GAIN: track.replayGain.toString(),
                REPLAYGAIN_TRACK_PEAK: track.peak.toString(),
                comment: track.url,
                isrc: track.isrc,
                upc: album.upc,
                artist: track.artists.map((a) => a.name),
                album: album.title,
                albumArtist: album.artists.map((a) => a.name),
                totalTracks: album.numberOfTracks.toString(),
                year: album.releaseDate.split("-")[0] || "",
            },
            picture: {
                buffer: Buffer.from(await (await fetch(`https://resources.tidal.com/images/${album.cover.replaceAll("-", "/")}/1280x1280.jpg`)).arrayBuffer()),
            }
        }).toBuffer()
    }
}